/*
								+---------------------------------+
								|                                 |
								|       ***   Directx   ***       |
								|                                 |
								|    Copyright  L. Polok  2001   |
								|                                 |
								|           Directx.cpp           |
								|                                 |
								+---------------------------------+
*/

#include <ddraw.h> 
#include "directx.h"

#define Depth 32
//#define DEBUG

static int n_Width, n_Height;
static int phase = 0;
static LPDIRECTDRAW         lpDD; 
static LPDIRECTDRAWSURFACE  lpDDSPrimary; 
static LPDIRECTDRAWSURFACE  lpDDSBack;
static LPDIRECTDRAWSURFACE  lpDDSOne;     
static LPDIRECTDRAWSURFACE  lpDDSTwo;
static DDSURFACEDESC        ddsd;
static unsigned __int32 *debugram;

HRESULT InitDirectX(HWND hwnd, int n_New_Width, int n_New_Height)
{
	n_Width = n_New_Width;
	n_Height = n_New_Height;

#ifdef DEBUG
	if((debugram = (unsigned __int32*) malloc(n_Width *
	   n_Height * sizeof(unsigned __int32))) == NULL)
		return -1;
#endif

#ifndef DEBUG
	// Inicializace DDraw ...
	HRESULT ddrval;
	DDSCAPS ddscaps;

	ddrval = DirectDrawCreate(NULL, &lpDD, NULL);
	if(ddrval == DD_OK)	{
		ddrval = lpDD->SetCooperativeLevel(hwnd, 
			DDSCL_EXCLUSIVE | DDSCL_ALLOWREBOOT | DDSCL_FULLSCREEN);
		if(ddrval == DD_OK) {
			ddrval = lpDD->SetDisplayMode(n_Width, n_Height, Depth);
			if(ddrval != DD_OK)
				return ddrval;
		} else return ddrval;
	} else return ddrval;
	// Vytvoreno zakladni ddraw objekt

	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP |
                          DDSCAPS_COMPLEX | DDSCAPS_HWCODEC;
    ddsd.dwBackBufferCount = 1;
    ddrval = lpDD->CreateSurface(&ddsd, &lpDDSPrimary, NULL);
	if(ddrval != DD_OK)
        return ddrval;

    ddscaps.dwCaps = DDSCAPS_BACKBUFFER | DDSCAPS_HWCODEC;
    ddrval = lpDDSPrimary->GetAttachedSurface(&ddscaps, &lpDDSBack);
    if(ddrval != DD_OK)
        return ddrval;
    // Vytvor offscreen bitmapy

    ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY
		| DDSCAPS_HWCODEC;
    ddsd.dwHeight = n_Height;
    ddsd.dwWidth = n_Width;
    ddrval = lpDD->CreateSurface(&ddsd, &lpDDSOne, NULL); // Prvni
    if(ddrval != DD_OK)
        return ddrval;
    ddrval = lpDD->CreateSurface(&ddsd, &lpDDSTwo, NULL); // A druha
    if(ddrval != DD_OK)
        return ddrval;
#endif
	// Yahoo !!!
	return DD_OK;
}

void TerminateDirectX(void)
{
#ifdef DEBUG
	free((void*)debugram);
#endif
#ifndef DEBUG
	// Znic DDraw objects
    if( lpDD != NULL )
    {
        if( lpDDSPrimary != NULL )
        {
            lpDDSPrimary->Release();
            lpDDSPrimary = NULL;
        }
        if( lpDDSOne != NULL )
        {
            lpDDSOne->Release();
            lpDDSOne = NULL;
        }
        if( lpDDSTwo != NULL )
        {
            lpDDSTwo->Release();
            lpDDSTwo = NULL;
        }
        lpDD->Release();
        lpDD = NULL;
    }
#endif
} 

HRESULT restoreAll(void)
{
#ifndef DEBUG
	HRESULT     ddrval;
	// Restore vsechny ddraw objects (chba pri kresleni)...
    ddrval = lpDDSPrimary->Restore();
    if( ddrval == DD_OK )
    {
        ddrval = lpDDSOne->Restore();
        if( ddrval == DD_OK )
        {
            ddrval = lpDDSTwo->Restore();
        }
    }
    return ddrval;
#else
	return DD_OK;
#endif
} 

int InitVideoBuffer(unsigned __int32 **pole)
{
#ifdef DEBUG
	*pole = debugram;
#endif
#ifndef DEBUG
	ddsd.dwSize = sizeof(ddsd);                  // Velikost bufferu
	if(phase){                                   // Nacti Surface, ktere ma byt aktivni
		if (lpDDSOne->Lock(NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL) != DD_OK)
				return 0;
		if (ddsd.ddpfPixelFormat.dwRGBBitCount != Depth)
		{
			lpDDSOne->Unlock(NULL);
			return 0;
		}
	}
	if(!phase){
		if (lpDDSTwo->Lock(NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL) != DD_OK)
				return 0;
		if (ddsd.ddpfPixelFormat.dwRGBBitCount != Depth)
		{
			lpDDSTwo->Unlock(NULL);
			return 0;
		}
	}
	*pole = (unsigned __int32 *)ddsd.lpSurface;   // Nacteni video RAM
#endif
	return 1;
}

void FlipVideoBuffer(void)
{
	// Init !!!
#ifndef DEBUG
	HRESULT ddrval;
	LPDIRECTDRAWSURFACE pdds;
	RECT rcRect;
	rcRect.left = 0;
    rcRect.top = 0;
    rcRect.right = n_Width;
    rcRect.bottom = n_Height;
	if(phase){
		phase = 0;
		lpDDSOne->Unlock(NULL);
		pdds = lpDDSOne;
	}		
	else{
		phase = 1;
		lpDDSTwo->Unlock(NULL);
		pdds = lpDDSTwo;
	}
	// Blt FAST !!!!
	while( 1 )
    {
        ddrval = lpDDSBack->BltFast( 0, 0, pdds, &rcRect, FALSE );
        if( ddrval == DD_OK )
        {
            break;
        }
        if( ddrval == DDERR_SURFACELOST )
        {
            ddrval = restoreAll();
            if( ddrval != DD_OK )
            {
                break;
            }
        }
        if( ddrval != DDERR_WASSTILLDRAWING )
        {
            break;
        }
    }
	// Flip Surfaces !!!
	while(1)
    {
        ddrval = lpDDSPrimary->Flip( NULL, 0 );
        if( ddrval == DD_OK )
        {
            break;
        }
        if( ddrval == DDERR_SURFACELOST )
        {
            ddrval = restoreAll();
            if( ddrval != DD_OK )
            {
                break;
            }
        }
        if( ddrval != DDERR_WASSTILLDRAWING )
        {
            break;
        }
    }
#endif
}


char * err_to_string(HRESULT h)
{
	switch(h) {
	case (DDERR_ALREADYINITIALIZED):
		return "Objekt nebyl spn inicializovn !";
		break;
	case (DDERR_BLTFASTCANTCLIP):
		return "Chyba funkce IDirectDrawSurface2::BltFast !";
		break;
	case (DDERR_CANNOTATTACHSURFACE):
		return "Chyba pipojen objektu IDirectDrawSurface2 !";
		break;
	case (DDERR_CANNOTDETACHSURFACE):
		return "Chyba odpojen objektu IDirectDrawSurface2  !";
		break;
	case (DDERR_CANTCREATEDC):
		return "Chyba vytvoen objektu HDc !";
		break;
	case (DDERR_CANTDUPLICATE):
		return "Chyba duplikace surface !";
		break;
	case (DDERR_CANTLOCKSURFACE):
		return "Objekt Surface nelze uzamknout !";
		break;
	case (DDERR_CANTPAGELOCK):
		return "Objekt Page nelze uzamknout !";
		break;
	case (DDERR_CLIPPERISUSINGHWND):
		return "Objekt DirectDrawClipper prv vyuv tda HWND!";
		break;
	case (DDERR_COLORKEYNOTSET):
		return "Pro operaci nebyl specifikovn ColorKey !";
		break;
	case (DDERR_CURRENTLYNOTAVAIL):
		return "Nepodporovan akce !";
		break;
	case (DDERR_DCALREADYCREATED):
		return "Objekt HDc nemohl bt vytvoen !";
		break;
	case (DDERR_DIRECTDRAWALREADYCREATED):
		return "Objekt DirectDraw nemohl bt vytvoen !";
		break;
	case (DDERR_EXCEPTION):
		return "Vjimen stav !!!";
		break;
	case (DDERR_EXCLUSIVEMODEALREADYSET):
		return "Cooperative mode nemohl bt nastaven na EXCLUSIVE !";
		break;
	case (DDERR_GENERIC):
		return "Nedefinovan chyba !";
		break;
	case (DDERR_HEIGHTALIGN):
		return "Height align !";
		break;
	case (DDERR_HWNDALREADYSET):
		return "Cooperative mode nemohl bt nastaven !";
		break;
	case (DDERR_HWNDSUBCLASSED):
		return "Hwnd bylo Subclassed !";
		break;
	case (DDERR_IMPLICITLYCREATED):
		return "Chyba implicitn vytvoenho surface !";
		break;
	case (DDERR_INCOMPATIBLEPRIMARY ):
		return "Nekompatibiln primary surface !";
		break;
	case (DDERR_INVALIDCAPS):
		return "Invalid caps !";
		break;
	case (DDERR_INVALIDCLIPLIST):
		return "Invalid cliplist !";
		break;
	case (DDERR_INVALIDDIRECTDRAWGUID):
		return "patn DDraw GUID (verze) !";
		break;
	case (DDERR_INVALIDMODE):
		return "DirectDraw nepodporuje tento md !";
		break;
	case (DDERR_INVALIDOBJECT):
		return "Pouit patnho DDraw objektu !";
		break;
	case (DDERR_INVALIDPARAMS):
		return "patn parametry !";
		break;
	case (DDERR_INVALIDPIXELFORMAT):
		return "patn PixelFormat !";
		break;
	case (DDERR_INVALIDPOSITION):
		return "patn pozice (peteen) !";
		break;
	case (DDERR_INVALIDRECT):
		return "patn pravohelnk !";
		break;
	case (DDERR_INVALIDSURFACETYPE):
		return "patn typ surface !";
		break;
	case (DDERR_LOCKEDSURFACES):
		return "Surface je uzamen !";
		break;
	case (DDERR_NO3D):
		return "Nen prezentovno hardwarov 3D !";
		break;
	case (DDERR_NOALPHAHW):
		return "Nen prezentovno alpha hardware urychlen !";
		break;
	case (DDERR_NOBLTHW):
		return "Nen prezentovn hardware Blitting !";
		break;
	case (DDERR_NOCLIPLIST):
		return "Clip list nen dostupn !";
		break;
	case (DDERR_NOCLIPPERATTACHED):
		return "Clipper nen pipojen k surface !";
		break;
	case (DDERR_NOCOLORCONVHW):
		return "Tato operace vyaduje hardware pro barevnou konverzi !";
		break;
	case (DDERR_NOCOLORKEY):
		return "Objekt surface nem sprvn ColorKey !";
		break;
	case (DDERR_NOCOLORKEYHW):
		return "Operace nelze provst bez hardware podporujcho ColorKey !";
		break;
	case (DDERR_NOCOOPERATIVELEVELSET):
		return "Nenastaveno CooperativeLevel !";
		break;
	case (DDERR_NODC):
		return "Toto Surface neme mt HDc !";
		break;
	case (DDERR_NODDROPSHW):
		return "Hardware nepodporuje DirectDrawRops !";
		break;
	case (DDERR_NODIRECTDRAWHW):
		return "Hardware nepodporuje DDraw !";
		break;
	case (DDERR_NODIRECTDRAWSUPPORT):
		return "Driver nepodporuje DDraw !";
		break;
	case (DDERR_NOEMULATION):
		return "Nenalezen emulan software !";
		break;
	case (DDERR_NOEXCLUSIVEMODE):
		return "Tato operace vyaduje EXCLUSIVE mode !";
		break;
	case (DDERR_NOFLIPHW):
		return "Hardware nepodporuje Surface flipping !";
		break;
	case (DDERR_NOGDI):
		return "Nen prezentovno GDI !";
		break;
	case (DDERR_NOHWND):
		return "Tato funkce potebuje sprvn Hwnd !";
		break;
	case (DDERR_NOMIPMAPHW):
		return "Hardware nepodporuje Mip-Mapping !";
		break;
	case (DDERR_NOMIRRORHW):
		return "Hardware nepodporuje Mirrors !";
		break;
	case (DDERR_NOOVERLAYDEST):
		return "Chyba pi voln ::GetOverlayPosition nebo ::UpdateOverlay !";
		break;
	case (DDERR_NOOVERLAYHW):
		return "Hardware nepodporuje Overlay !";
		break;
	case (DDERR_NOPALETTEATTACHED):
		return "K Surface nebyla pipojena paleta !";
		break;
	case (DDERR_NOPALETTEHW):
		return "Hardware nepodporuje 16(256) - barevn palety !";
		break;
	case (DDERR_NORASTEROPHW):
		return "Neuskuteniteln hardwarov raster operace !";
		break;
	case (DDERR_NOROTATIONHW):
		return "Hardware nepodporuje Rotace !";
		break;
	case (DDERR_NOSTRETCHHW):
		return "Hardware nepodporuje Stretching !";
		break;
	case (DDERR_NOT4BITCOLOR):
		return "Operace poaduje 4-Bit paletu !";
		break;
	case (DDERR_NOT4BITCOLORINDEX):
		return "Operace vyaduje 4-Bit barevn index !";
		break;
	case (DDERR_NOT8BITCOLOR):
		return "Operace poaduje 8-Bit paletu !";
		break;
	case (DDERR_NOTAOVERLAYSURFACE):
		return "Overlay komponent voln z non-overlay surface !";
		break;
	case (DDERR_NOTEXTUREHW):
		return "Hardware nepodporuje texturing !";
		break;
	case (DDERR_NOTFLIPPABLE):
		return "Surface se nepodailo FLIPovat !";
		break;
	case (DDERR_NOTFOUND):
		return "Poadovan soust nenalezena !";
		break;
	case (DDERR_NOTINITIALIZED):
		return "Poadovan soust neinicializovna !";
		break;
	case (DDERR_NOTLOCKED):
		return "Surface nen zamen (a proto nelze odemt) !";
		break;
	case (DDERR_NOTPAGELOCKED):
		return "Page nen zamen (a proto nelze odemt) !";
		break;
	case (DDERR_NOTPALETTIZED):
		return "Surface nen Palette-based !";
		break;
	case (DDERR_NOVSYNCHW):
		return "Hardware nepodporuje vertikln synchronizaci !";
		break;
	case (DDERR_NOZBUFFERHW):
		return "Hardware nepodporuje Z-Buffer !";
		break;
	case (DDERR_NOZOVERLAYHW):
		return "Hardware nepodporuje Z-Overlay !";
		break;
	case (DDERR_OUTOFCAPS):
		return "Operace potebuje vce fyzick pamti pro alokaci !";
		break;
	case (DDERR_OUTOFMEMORY):
		return "Mlo RAM !";
		break;
	case (DDERR_OUTOFVIDEOMEMORY):
		return "Mlo Video-RAM ! !";
		break;
	case (DDERR_OVERLAYCANTCLIP):
		return "Hardware nepodporuje Clipped Overlays !";
		break;
	case (DDERR_OVERLAYCOLORKEYONLYONEACTIVE):
		return "Na overlay je aktivn jen 1 ColorKey !";
		break;
	case (DDERR_OVERLAYNOTVISIBLE):
		return "Metoda IDirectDrawSurface2::GetOverlayPosition je volan pro skryt overlay !";
		break;
	case (DDERR_PALETTEBUSY):
		return "V tto chvli je paleta nepstupn (zkuste to za 10 min.) !";
		break;
	case (DDERR_PRIMARYSURFACEALREADYEXISTS):
		return "Tento proces vyaduje primary surface !";
		break;
	case (DDERR_REGIONTOOSMALL):
		return "Region zadan funkci IDirectDrawClipper::GetClipList  je pli mal !";
		break;
	case (DDERR_SURFACEALREADYATTACHED):
		return "Surface nelze napojit !";
		break;
	case (DDERR_SURFACEALREADYDEPENDENT):
		return "Surface napojen na jinou neme bt zvisl !";
		break;
	case (DDERR_SURFACEBUSY):
		return "Surface nyn nen k dispozici (zkuste to za 10 min.) !";
		break;
	case (DDERR_SURFACEISOBSCURED):
		return "Pstup k surface byl odmtnut !";
		break;
	case (DDERR_SURFACELOST ):
		return "Pstup k surface byl ztracen !";
		break;
	case (DDERR_SURFACENOTATTACHED):
		return "Poadovan surface nen zvisl !";
		break;
	case (DDERR_TOOBIGHEIGHT):
		return "Height je pli velk !";
		break;
	case (DDERR_TOOBIGSIZE):
		return "Velikost je pli velk !";
		break;
	case (DDERR_TOOBIGWIDTH):
		return "Width je pli velk !";
		break;
	case (DDERR_UNSUPPORTED):
		return "Operace nen podporovna !";
		break;
	case (DDERR_UNSUPPORTEDFORMAT):
		return "Formt nen podporovn !";
		break;
	case (DDERR_UNSUPPORTEDMASK ):
		return "BitMask nen podporovna !";
		break;
	case (DDERR_UNSUPPORTEDMODE):
		return "Mode nen podporovn !";
		break;
	case (DDERR_VERTICALBLANKINPROGRESS):
		return "Vertikln msto se zpracovv !";
		break;
	case (DDERR_WASSTILLDRAWING):
		return "Ped BLIT se pi transferu stala chyba !";
		break;
	case (DDERR_WRONGMODE):
		return "patn md !";
		break;
	case (DDERR_XALIGN):
		return "Pravohelnk neme bt horizontln zarovnn !";
		break;
	default:
		return "Neidentifikovateln chyba !";
		break;
		}
}

void BuffSet(unsigned __int32 *vid, __int32 val, long cnt)
{
	int i;
	for(i = 0;i < cnt; i ++)
		vid[i] = val;
}

