/*
								+---------------------------------+
								|                                 |
								|   ***   Segment_Manager   ***   |
								|                                 |
								|    Copyright  L. Polok  2002   |
								|                                 |
								|           Seg_Man.cpp           |
								|                                 |
								+---------------------------------+
*/

#include <stdlib.h>
#include <string.h>
#include "Load_3ds.h"
#include "Seg_Man.h"

#define SEG_SPLIT(a,b) (((a)->dzdx == (b)->dzdx)? -1 : \
	(int)(0.5 + ((a)->z1 - (b)->z1) / ((b)->dzdx - (a)->dzdx)))
#define SEG_Z(a,b) ((a)->z1 + (a)->dzdx * b)

static int startup_seg = 1024;
static int seg_inc     = 256;

static int  mem_stat;
static int  stck_num;
static int stck_used;
static int  n_seg_used;
static int   n_seg_num;
static Segment **stack;

// inits Segment manager (fast memory allocator)
int InitSManager()
{
	int i;

	stck_num  = 0;
	n_seg_num   = startup_seg;
	stck_used = 0;
	n_seg_used  = 0;
	// setup status

	mem_stat = Mem_OK; 
	// memory allocation status = ok

	if((stack = (Segment**) malloc((stck_num + 1) * sizeof(Segment*))) == NULL)
		return 0;
	if((stack[0] = (Segment*) malloc(n_seg_num * sizeof(Segment))) == NULL)
		return 0;
	for(i = 1; i < stck_num; i ++){
		if((stack[i] = (Segment*) malloc(seg_inc * sizeof(Segment))) == NULL)
			return 0;
	}
	// alloc memory for Segments

	return 1;
}

// uninits Segment manager
void UninitSManager()
{
	int i;

	for(i = 0; i < (stck_num + 1); i ++)
		free((void*)stack[i]);
	free((void*)stack);
	// freeing stack memory

	stack = NULL;
	stck_num  = 0;
	n_seg_num   = 0;
	stck_used = 0;
	n_seg_used  = 0;
	// sets segman to uninti
}

// returns current amount of used memory
int GetMem_Usage()
{
	return (((stck_num * seg_inc)
		+ startup_seg) * sizeof(Segment));
}

// return current status of manager (ok / failed)
int GetMem_Status()
{
	return mem_stat;
}

// set all Segments to "unused"
void ResetSManager()
{
	stck_used = 0;
	n_seg_used  = 0;
}

// grovs Segment reservoir
int GrowReservoir()
{
	stck_num ++;
	n_seg_num = seg_inc;
	n_seg_used = 0;
	mem_stat = Mem_Insufficient;

	if((stack = (Segment**) realloc(stack, 
	   (stck_num + 1) * sizeof(Segment*))) == NULL)
		return 0;
	if((stack[stck_num] = (Segment*) malloc(
	   seg_inc * sizeof(Segment))) == NULL)
		return 0;

	mem_stat = Mem_OK;

	return 1;
}

// inits Scanline to have seg_n free Segments
int InitScanline(int seg_n, Scanline *sl)
{
    Segment *act;
	int i;

	sl->n_seg_used = 0;
	sl->n_seg_num = 1;

	if(n_seg_used == n_seg_num) {
		if(stck_used >= stck_num){
			if(!GrowReservoir())
				return 0;
		}
		stck_used ++;
		n_seg_used = 0;
	}
	sl->segm = &stack[stck_used][n_seg_used];
	n_seg_used ++;
	sl->last = sl->segm;
	sl->segm->prev = NULL;
	sl->segm->next = NULL;
	act = sl->segm;
	// first Segment

	for(i = 1; i < seg_n; i ++) {
		if(n_seg_used == n_seg_num) {
			if(stck_used >= stck_num){
				if(!GrowReservoir())
					return 0;
			}
			stck_used ++;
			n_seg_used = 0;
		}
		act->next = &stack[stck_used][n_seg_used];
		act->next->prev = act;
		act->next->next = NULL;
		act = act->next;
		sl->last = act;
		
		sl->n_seg_num ++;
		n_seg_used ++;
	}
	// Next Segments

	sl->n_seg_used = 0;
	sl->last->next = NULL;

	return 1;
}

// gets one Segment(last) from Scanline or if there
// are no unused Segments, add some into Scanline
inline Segment *Get_Segment(Scanline *sl, Segment *seg)
{
	int i;
    Segment *s;

	if(n_seg_used == n_seg_num) {
		if(stck_used >= stck_num){
			if(!GrowReservoir())
				return NULL;
		}
		stck_used ++;
		n_seg_used = 0;
	}
	if(sl->n_seg_used == sl->n_seg_num) {
		sl->last->next = &(stack[stck_used][n_seg_used]);
		memcpy(sl->last->next, seg, sizeof(Segment));
		sl->last->next->prev = sl->last;
		sl->last = sl->last->next;
		sl->last->next = NULL;
		sl->n_seg_num ++;
		n_seg_used ++;
		sl->n_seg_used ++;

		return sl->last;
	}

	s = sl->segm;
	for(i = 0; i < sl->n_seg_used; i ++)
		s = s->next;
	sl->n_seg_used ++;

	seg->next = s->next;
	seg->prev = s->prev;

	memcpy(s, seg, sizeof(Segment));

	return s;
}

// gets one seg, lying behind cur from Scanline 
inline Segment *GetSegmentBehind(Scanline *sl, Segment *cur, Segment *seg)
{
	Segment *s;

	if(n_seg_used == n_seg_num) {
		if(stck_used == stck_num) {
			if(!GrowReservoir())
				return NULL;
		}
		stck_used ++;
		n_seg_used = 0;
	}
	if(sl->n_seg_used == sl->n_seg_num) {
		s = &(stack[stck_used][n_seg_used]);
		memcpy(s, seg, sizeof(Segment));
		if(cur->prev != NULL)
			cur->prev->next = s;
		else
			sl->segm = s; // Segment become first
		s->prev = cur->prev;
		s->next = cur;
		cur->prev = s;
		sl->n_seg_num ++;
		sl->n_seg_used ++;
		n_seg_used ++;

		return s;
	}
	
	s = sl->last;

	sl->last = s->prev;
	sl->last->next = NULL;
	
	memcpy(s, seg, sizeof(Segment));

	if(cur->prev != NULL)
		cur->prev->next = s;
	else
		sl->segm = s; // Segment become first
	s->prev = cur->prev;
	s->next = cur;
	cur->prev = s;

	sl->n_seg_used ++;
	
	return s;
}

// gets one seg, lying next to cur from Scanline 
inline Segment *GetSegmentNext(Scanline *sl, Segment *cur, Segment *seg)
{
	Segment *s;

	if(n_seg_used == n_seg_num) {
		if(stck_used >= stck_num){
			if(!GrowReservoir())
				return NULL;
		}
		stck_used ++;
		n_seg_used = 0;
	}
	if(sl->n_seg_used == sl->n_seg_num){
		s = &(stack[stck_used][n_seg_used]);
		memcpy(s, seg, sizeof(Segment));
		if(cur->next != NULL)
			cur->next->prev = s;
		else
			sl->last = s; // Segment become last
		s->prev = cur;
		s->next = cur->next;
		cur->next = s;
		sl->n_seg_num ++;
		sl->n_seg_used ++;
		n_seg_used ++;

		return s;
	}

	s = sl->last;
	
	sl->last = sl->last->prev;
	sl->last->next = NULL;

	memcpy(s, seg, sizeof(Segment));

	if(cur->next != NULL)
		cur->next->prev = s;
	else
		sl->last = s; // Segment become last
	s->prev = cur;
	s->next = cur->next;
	cur->next = s;

	sl->n_seg_used ++;
	
	return s;
}

// removes Segment from Scanline
inline void RemoveSegment(Scanline *sl, Segment *seg)
{
	if(seg->next == NULL) { // Segment is on the end
		sl->n_seg_used --;

		return;
	}
	if(seg->prev == NULL) { // Segment is on the beginning
		sl->segm = seg->next;
		seg->next->prev = NULL;

		sl->last->next = seg;
		seg->prev = sl->last;
		seg->next = NULL;
		sl->last = sl->last->next;

		sl->n_seg_used --;

		return;
	}

	// Segment is anywhere inside of buffer
	seg->next->prev = seg->prev;
	seg->prev->next = seg->next;

	sl->last->next = seg;
	seg->prev = sl->last;
	seg->next = NULL;
	sl->last = sl->last->next;

	sl->n_seg_used --;
}

// adds Segment to Scanline (core fuction) and
// makes intersection with all other Segments
void _fastcall AddSegment(Segment *seg, Scanline *sc)
{
	Segment *tmp, *cur;
	int i = 0, x, s;

	cur = sc->segm;
	while(i < sc->n_seg_used) {
		if(seg->x2 <= cur->x1) { // Case # 1
			GetSegmentBehind(sc, cur, seg);
			return;
		} else if(seg->x1 >= cur->x2) { // Case # 2
			i ++;
			cur = cur->next;
			continue;
		} else if(seg->x2 > cur->x1 && seg->x1 < cur->x1 && seg->x2 < cur->x2) { // Case # 3
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					cur->x1 = x;
					seg->x2 = x;
					GetSegmentBehind(sc, cur, seg);
					return;
				} else {
					s = cur->x1;
					cur->x1 = seg->x2;
					if((tmp = GetSegmentBehind(sc, cur, seg)) == NULL)
						return;
					tmp->x1 = x;
					//cur = tmp;
					//cur = cur->next;
					if((tmp = GetSegmentBehind(sc, tmp/*cur*/, cur)) == NULL)
						return;
					tmp->x1 = s;
					tmp->x2 = x;
					if((tmp = GetSegmentBehind(sc, tmp/*cur*/, seg)) == NULL)
						return;
					tmp->x2 = s;
					return;
				}
			} else {
				x = cur->x1 + (int)((float)(seg->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					cur->x1 = seg->x2;
					GetSegmentBehind(sc, cur, seg);
					return;
				} else {
					seg->x2 = cur->x1;
					GetSegmentBehind(sc, cur, seg);
					return;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x1 < cur->x2 && seg->x2 > cur->x2) { // Case # 4
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					s = cur->x2;
					cur->x2 = seg->x1;
					if((tmp = GetSegmentNext(sc, cur, seg)) == NULL)
						return;
					tmp->x2 = x;
					if((tmp = GetSegmentNext(sc, tmp, cur)) == NULL)
						return;
					tmp->x1 = x;
					tmp->x2 = s;
					seg->x1 = s;
					cur = tmp;
					i += 2;
					continue;
				} else {
					cur->x2 = x;
					seg->x1 = x;
					cur = cur->next;
					i ++;
					continue;
				}
			} else {
				x = seg->x1 + (int)((float)(cur->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					cur->x2 = seg->x1;
					i ++;
					cur = cur->next;
					continue;
				} else {
					seg->x1 = cur->x2;
					i ++;
					cur = cur->next;
					continue;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x2 < cur->x2) { // Case # 5
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					if((tmp = GetSegmentNext(sc, cur, seg)) == NULL)
						return;
					if((tmp = GetSegmentNext(sc, cur->next, cur)) == NULL)
						return;
					tmp->x1 = x;
					cur->x2 = seg->x1;
					return;
				} else {
					seg->x1 = x;
					if((tmp = GetSegmentNext(sc, cur, seg)) == NULL)
						return;
					if((tmp = GetSegmentNext(sc, cur->next, cur)) == NULL)
						return;
					tmp->x1 = seg->x2;
					cur->x2 = x;
					return;
				}
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					if((tmp = GetSegmentNext(sc, cur, seg)) == NULL)
						return;
					if((tmp = GetSegmentNext(sc, cur->next, cur)) == NULL)
						return;
					tmp->x1 = seg->x2;
					cur->x2 = seg->x1;
					return;
				}
				return;
			}
		} else if(seg->x1 < cur->x1 && seg->x2 > cur->x2) { // Case # 6
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					s = cur->x2;
					cur->x1 = x;
					if((tmp = GetSegmentBehind(sc, cur, seg)) == NULL)
						return;
					tmp->x2 = x;
					seg->x1 = s;
					cur = cur->next;
					i += 2;
					continue;
				} else {
					s = cur->x1;
					cur->x2 = x;
					if((tmp = GetSegmentBehind(sc, cur, seg)) == NULL)
						return;
					tmp->x2 = s;
					seg->x1 = x;
					cur = cur->next;
					i += 2;
					continue;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					tmp = cur;
					cur = cur->next;
					RemoveSegment(sc, tmp);
					continue;
				} else {
					x = cur->x1;
					s = cur->x2;
					if((tmp = GetSegmentBehind(sc, cur, seg)) == NULL)
						return;
					tmp->x2 = x;
					seg->x1 = s;
					cur = cur->next;
					i += 2;
					continue;
				}
			}
		} else if(seg->x1 == cur->x1 && seg->x2 < cur->x2) { // Case # 7a
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					cur->x1 = x;
					GetSegmentBehind(sc, cur, seg);
					return;
				} else {
					s = cur->x2;
					cur->x2 = x;
					if((tmp = GetSegmentNext(sc, cur, seg)) == NULL)
						return;
					tmp->x1 = x;
					if((tmp = GetSegmentNext(sc, cur->next, cur)) == NULL)
						return;
					tmp->x1 = seg->x2;
					tmp->x2 = s;
					return;
				}
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					cur->x1 = seg->x2;
					GetSegmentBehind(sc, cur, seg);
					return;
				}
				return;
			}
		} else if(seg->x1 == cur->x1 && seg->x2 > cur->x2) { // Case # 7b
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					s = cur->x2;
					cur->x1 = x;
					if((tmp = GetSegmentBehind(sc, cur, seg)) == NULL)
						return;
					tmp->x2 = x;
					seg->x1 = s;
					cur = cur->next;
					i += 2;
					continue;
				} else {
					cur->x2 = x;
					seg->x1 = x;
					i ++;
					cur = cur->next;
					continue;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					tmp = cur;
					cur = cur->next;
					RemoveSegment(sc, tmp);
					continue;
				} else {
					seg->x1 = cur->x2;
					i ++;
					cur = cur->next;
					continue;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x2 == cur->x2) { // Case # 8a
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					s = cur->x2;
					cur->x2 = seg->x1;
					seg->x2 = x;
					if((tmp = GetSegmentNext(sc, cur, seg)) == NULL)
						return;
					if((tmp = GetSegmentNext(sc, cur->next, cur)) == NULL)
						return;
					tmp->x1 = x;
					tmp->x2 = s;
					return;
				} else {
					cur->x2 = x;
					seg->x1 = x;
					GetSegmentNext(sc, cur, seg);
					return;
				}
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					cur->x2 = seg->x1;
					GetSegmentNext(sc, cur, seg);
					return;
				}
				return;
			}
		} else if(seg->x1 < cur->x1 && seg->x2 == cur->x2) { // Case # 8b
			x = SEG_SPLIT(cur, seg);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					cur->x1 = x;
					GetSegmentBehind(sc, cur, seg);
					return;
				} else {
					cur->x2 = x;
					if((tmp = GetSegmentNext(sc, cur, seg)) == NULL)
						return;
					tmp->x1 = x;
					seg->x2 = cur->x1;
					GetSegmentBehind(sc, cur, seg);
					return;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					seg->prev = cur->prev;
					seg->next = cur->next;
					memcpy(cur, seg, sizeof(Segment));
					return;
				} else {
					seg->x2 = cur->x1;
					GetSegmentBehind(sc, cur, seg);
					return;
				}
			}
		} else if(seg->x1 == cur->x1 && seg->x2 == cur->x2) { // Case # 9
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					cur->x1 = x;
					GetSegmentBehind(sc, cur, seg);
					return;
				} else {
					seg->x1 = x;
					cur->x2 = x;
					GetSegmentNext(sc, cur, seg);
					return;
				}
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					seg->prev = cur->prev;
					seg->next = cur->next;
					memcpy(cur, seg, sizeof(Segment));
					return;
				}
				return;
			}
		}
		
		cur = cur->next;
		i ++;
	}

	if(Get_Segment(sc, seg) == NULL)
		return;
}

int _fastcall Segment_Visibility(Segment *seg, Scanline *sc)
{
	Segment *cur;
	int i = 0, x;
	int res = 0;

	cur = sc->segm;
	while(i < sc->n_seg_used) {
		if(seg->x2 <= cur->x1) { // Case # 1
			return (res + seg->x2 - seg->x1);
		} else if(seg->x1 >= cur->x2) { // Case # 2
			i ++;
			cur = cur->next;
			continue;
		} else if(seg->x2 > cur->x1 && seg->x1 < cur->x1 && seg->x2 < cur->x2) { // Case # 3
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return (res + x - seg->x1);
				else
					return (res + seg->x2 - x + cur->x1 - seg->x1);
			} else {
				x = cur->x1 + (int)((float)(seg->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return (res + seg->x2 - seg->x1);
				else
					return (res + cur->x1 - seg->x1);
			}
		} else if(seg->x1 > cur->x1 && seg->x1 < cur->x2 && seg->x2 > cur->x2) { // Case # 4
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					res += seg->x1 - x;
					seg->x1 = cur->x2;
					cur = cur->next->next;
					i += 2;
					continue;
				} else {
					seg->x1 = x;
					cur = cur->next;
					i ++;
					continue;
				}
			} else {
				x = seg->x1 + (int)((float)(cur->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					i ++;
					cur = cur->next;
					continue;
				} else {
					seg->x1 = cur->x2;
					i ++;
					cur = cur->next;
					continue;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x2 < cur->x2) { // Case # 5
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return (res + x - seg->x1);
				else
					return (res + seg->x2 - x);
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					return (res + seg->x2 - seg->x1);
				}
				return res;
			}
		} else if(seg->x1 < cur->x1 && seg->x2 > cur->x2) { // Case # 6
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					res += x - seg->x1;
					seg->x1 = cur->x2;
					cur = cur->next;
					i += 2;
					continue;
				} else {
					res += cur->x2 - seg->x1;
					cur = cur->next;
					i += 2;
					continue;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					cur = cur->next;
					i ++;
					continue;
				} else {
					res += cur->x1 - seg->x1;
					seg->x1 = cur->x2;
					cur = cur->next;
					i += 2;
					continue;
				}
			}
		} else if(seg->x1 == cur->x1 && seg->x2 < cur->x2) { // Case # 7a
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return (res + x - seg->x1);
				else
					return (res + seg->x2 - x);
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return (res + seg->x2 - seg->x1);
				return res;
			}
		} else if(seg->x1 == cur->x1 && seg->x2 > cur->x2) { // Case # 7b
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					res += x - seg->x1;
					seg->x1 = cur->x2;
					cur = cur->next->next;
					i += 2;
					continue;
				} else {
					seg->x1 = x;
					i ++;
					cur = cur->next;
					continue;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					cur = cur->next;
					i ++;
					continue;
				} else {
					seg->x1 = cur->x2;
					i ++;
					cur = cur->next;
					continue;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x2 == cur->x2) { // Case # 8a
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return (res + x - seg->x1);
				else
					return (res + seg->x2 - x);
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return (res + seg->x2 - seg->x1);
				return res;
			}
		} else if(seg->x1 < cur->x1 && seg->x2 == cur->x2) { // Case # 8b
			x = SEG_SPLIT(cur, seg);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return (res + x - seg->x1);
				else
					return (res + seg->x2 - x);
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return (res + seg->x2 - seg->x1);
				else
					return (res + cur->x1 - seg->x1);
			}
		} else if(seg->x1 == cur->x1 && seg->x2 == cur->x2) { // Case # 9
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return (res + x - seg->x1);
				else
					return (res + seg->x2 - x);
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return (res + seg->x2 - seg->x1);
				return res;
			}
		}
		cur = cur->next;
		i ++;
	}

	return res;
}

int _fastcall Segment_Visible(Segment *seg, Scanline *sc)
{
	Segment *cur;
	int i = 0, x;

	cur = sc->segm;
	while(i < sc->n_seg_used) {
		if(seg->x2 <= cur->x1) { // Case # 1
			return 1;
		} else if(seg->x1 >= cur->x2) { // Case # 2
			i ++;
			cur = cur->next;
			continue;
		} else if(seg->x2 > cur->x1 && seg->x1 < cur->x1 && seg->x2 < cur->x2) { // Case # 3
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return 1;
				else
					return 1;
			} else {
				x = cur->x1 + (int)((float)(seg->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return 1;
				else
					return 1;
			}
		} else if(seg->x1 > cur->x1 && seg->x1 < cur->x2 && seg->x2 > cur->x2) { // Case # 4
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					return 1;
				} else {
					seg->x1 = x;
					cur = cur->next;
					i ++;
					continue;
				}
			} else {
				x = seg->x1 + (int)((float)(cur->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					i ++;
					cur = cur->next;
					continue;
				} else {
					seg->x1 = cur->x2;
					i ++;
					cur = cur->next;
					continue;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x2 < cur->x2) { // Case # 5
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return 1;
				else
					return 1;
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					return 1;
				}
				return 0;
			}
		} else if(seg->x1 < cur->x1 && seg->x2 > cur->x2) { // Case # 6
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					return 1;
				} else {
					return 1;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					cur = cur->next;
					i ++;
					continue;
				} else {
					return 1;
				}
			}
		} else if(seg->x1 == cur->x1 && seg->x2 < cur->x2) { // Case # 7a
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return 1;
				else
					return 1;
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return 1;
				return 0;
			}
		} else if(seg->x1 == cur->x1 && seg->x2 > cur->x2) { // Case # 7b
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					return 1;
				} else {
					seg->x1 = x;
					i ++;
					cur = cur->next;
					continue;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					cur = cur->next;
					i ++;
					continue;
				} else {
					seg->x1 = cur->x2;
					i ++;
					cur = cur->next;
					continue;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x2 == cur->x2) { // Case # 8a
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return 1;
				else
					return 1;
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return 1;
				return 0;
			}
		} else if(seg->x1 < cur->x1 && seg->x2 == cur->x2) { // Case # 8b
			x = SEG_SPLIT(cur, seg);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return 1;
				else
					return 1;
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return 1;
				else
					return 1;
			}
		} else if(seg->x1 == cur->x1 && seg->x2 == cur->x2) { // Case # 9
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000))
					return 1;
				else
					return 1;
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x))
					return 1;
				return 0;
			}
		}
		cur = cur->next;
		i ++;
	}

	return 1;
}

void TransparentSegment(Segment *seg, Scanline *sc,
	void (__fastcall drawfunc)(Segment*))
{
	int i = 0, x, s;
	Segment *cur;

	cur = sc->segm;
	while(i < sc->n_seg_used) {
		if(seg->x2 <= cur->x1) { // Case # 1
			drawfunc(seg);
			return;
		} else if(seg->x1 >= cur->x2) { // Case # 2
			i ++;
			cur = cur->next;
			continue;
		} else if(seg->x2 > cur->x1 && seg->x1 < cur->x1 && seg->x2 < cur->x2) { // Case # 3
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					drawfunc(seg);
					return;
				} else {
					s = seg->x2;
					seg->x2 = cur->x1;
					drawfunc(seg);
					seg->x1 = x;
					seg->x2 = s;
					drawfunc(seg);
					return;
				}
			} else {
				x = cur->x1 + (int)((float)(seg->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					drawfunc(seg);
					return;
				} else {
					seg->x2 = cur->x1;
					drawfunc(seg);
					return;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x1 < cur->x2 && seg->x2 > cur->x2) { // Case # 4
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					s = seg->x2;
					seg->x2 = x;
					drawfunc(seg);
					seg->x1 = cur->x2;
					seg->x2 = s;
					cur = cur->next;
					i ++;
					continue;
				} else {
					seg->x1 = x;
					cur = cur->next;
					i ++;
					continue;
				}
			} else {
				x = seg->x1 + (int)((float)(cur->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					i ++;
					cur = cur->next;
					continue;
				} else {
					seg->x1 = cur->x2;
					i ++;
					cur = cur->next;
					continue;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x2 < cur->x2) { // Case # 5
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					drawfunc(seg);
					return;
				} else {
					seg->x1 = x;
					drawfunc(seg);
					return;
				}
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					drawfunc(seg);
					return;
				}
				return;
			}
		} else if(seg->x1 < cur->x1 && seg->x2 > cur->x2) { // Case # 6
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					s = seg->x2;
					seg->x2 = x;
					drawfunc(seg);
					seg->x1 = cur->x2;
					seg->x2 = s;
					cur = cur->next;
					i ++;
					continue;
				} else {
					s = seg->x2;
					seg->x2 = cur->x1;
					drawfunc(seg);
					seg->x2 = s;
					seg->x1 = x;
					cur = cur->next;
					i ++;
					continue;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					cur = cur->next;
					i ++;
					continue;
				} else {
					x = seg->x2;
					seg->x2 = cur->x1;
					drawfunc(seg);
					seg->x1 = cur->x2;
					seg->x2 = x;
					cur = cur->next;
					i ++;
					continue;
				}
			}
		} else if(seg->x1 == cur->x1 && seg->x2 < cur->x2) { // Case # 7a
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					drawfunc(seg);
					return;
				} else {
					seg->x1 = x;
					drawfunc(seg);
					return;
				}
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					drawfunc(seg);
					return;
				}
				return;
			}
		} else if(seg->x1 == cur->x1 && seg->x2 > cur->x2) { // Case # 7b
			x = SEG_SPLIT(seg, cur);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					s = seg->x2;
					seg->x2 = x;
					drawfunc(seg);
					seg->x2 = s;
					seg->x1 = cur->x2;
					cur = cur->next;
					i ++;
					continue;
				} else {
					seg->x1 = x;
					i ++;
					cur = cur->next;
					continue;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					i ++;
					cur = cur->next;
					continue;
				} else {
					seg->x1 = cur->x2;
					i ++;
					cur = cur->next;
					continue;
				}
			}
		} else if(seg->x1 > cur->x1 && seg->x2 == cur->x2) { // Case # 8a
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					drawfunc(seg);
					return;
				} else {
					seg->x1 = x;
					drawfunc(seg);
					return;
				}
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					drawfunc(seg);
					return;
				}
				return;
			}
		} else if(seg->x1 < cur->x1 && seg->x2 == cur->x2) { // Case # 8b
			x = SEG_SPLIT(cur, seg);
			if(x > cur->x1 && x < cur->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					drawfunc(seg);
					return;
				} else {
					s = seg->x1;
					seg->x1 = x;
					drawfunc(seg);
					//
					seg->x1 = s;
					seg->x2 = cur->x1;
					drawfunc(seg);
					return;
				}
			} else {
				x = cur->x1 + (int)((float)(cur->x2 - cur->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					drawfunc(seg);
					return;
				} else {
					seg->x2 = cur->x1;
					drawfunc(seg);
					return;
				}
			}
		} else if(seg->x1 == cur->x1 && seg->x2 == cur->x2) { // Case # 9
			x = SEG_SPLIT(seg, cur);
			if(x > seg->x1 && x < seg->x2) {
				if(SEG_Z(cur, -0x1000) < SEG_Z(seg, -0x1000)) {
					seg->x2 = x;
					drawfunc(seg);
					return;
				} else {
					seg->x1 = x;
					drawfunc(seg);
					return;
				}
			} else {
				x = seg->x1 + (int)((float)(seg->x2 - seg->x1) / 2 + 0.5);
				if(SEG_Z(cur, x) < SEG_Z(seg, x)) {
					drawfunc(seg);
					return;
				}
				return;
			}
		}
		
		cur = cur->next;
		i ++;
	}

	drawfunc(seg);
}