/*
								+---------------------------------+
								|                                 |
								|   ***   Silhouette tool   ***   |
								|                                 |
								|    Copyright  L. Polok  2003   |
								|                                 |
								|          Silhouette.cpp         |
								|                                 |
								+---------------------------------+
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "matrix.h"
#include "load_3ds.h"
#include "pyramid.h"
#include "raster.h"
#include "silhouette.h"

#define _Len(A) ((float)sqrt(A.x * A.x + A.y * A.y + A.z * A.z))
// spot dlku vektoru

#define _Normalize(A) {float s = _Len(A); A.x /= s; A.y /= s; A.z /= s; }
// vytvo jednotkov vektor

#define _Dot(A,B) (B.x * A.x + A.y * B.y)
// skalrn souin

#define _Cross(A,B,C) C.x = B.y * A.z - A.y * B.z; \
                     C.y = B.z * A.x - A.z * B.x; \
                     C.z = B.x * A.y - A.x * B.y;
// vektorov souin

int b_SameVertex(Vertex *p_vertex_a, Vertex *p_vertex_b)
{
	float f, d;

	f = (p_vertex_a->x - p_vertex_b->x);
	f *= f;
	d = (p_vertex_a->y - p_vertex_b->y);
	f += d * d;
	d = (p_vertex_a->z - p_vertex_b->z);
	// vzdlenost mezi vrcholy
	
	return ((f + d * d) < 0.01f);
}

void Find_FaceNeighbours(Object *p_object)
{
	int n_edge_a, n_edge_b;
	int n_edge_c, n_edge_d;
	int i, j, k, l;

	for(i = 0; i < p_object->n_face_num; i ++) {
		p_object->face[i].n_neighbour[0] = NULL;
		p_object->face[i].n_neighbour[1] = NULL;
		p_object->face[i].n_neighbour[2] = NULL;
	}

	for(i = 0; i < p_object->n_face_num; i ++) {
		for(n_edge_a = 0; n_edge_a < 3; n_edge_a ++) {
			n_edge_b = (1 + n_edge_a) % 3;
			l = (1 + n_edge_b) % 3;
			// indexy vertex hrany (a, b)
			for(j = i + 1; j < p_object->n_face_num; j ++) {
				for(n_edge_c = 0; n_edge_c < 3; n_edge_c ++) {
					n_edge_d = (n_edge_c + 1) % 3;
					// indexy vertex druh hrany (c, d)

					if((b_SameVertex(p_object->face[i].vertex[n_edge_a],
									p_object->face[j].vertex[n_edge_d]) &&
					   b_SameVertex(p_object->face[i].vertex[n_edge_b],
									p_object->face[j].vertex[n_edge_c])) ||
					   (b_SameVertex(p_object->face[i].vertex[n_edge_a],
									p_object->face[j].vertex[n_edge_c]) &&
					   b_SameVertex(p_object->face[i].vertex[n_edge_b],
									p_object->face[j].vertex[n_edge_d]))) {
						k = (n_edge_d + 1) % 3;
						if(p_object->face[i].normal.a * p_object->face[j].vertex[k]->x +
						   p_object->face[i].normal.b * p_object->face[j].vertex[k]->y +
						   p_object->face[i].normal.c * p_object->face[j].vertex[k]->z +
						   p_object->face[i].normal.d > -.01 &&
						   p_object->face[j].normal.a * p_object->face[i].vertex[l]->x +
						   p_object->face[j].normal.b * p_object->face[i].vertex[l]->y +
						   p_object->face[j].normal.c * p_object->face[i].vertex[l]->z +
						   p_object->face[j].normal.d > -.01) {
							p_object->face[i].n_neighbour[n_edge_a] = &p_object->face[j];
							p_object->face[j].n_neighbour[n_edge_c] = &p_object->face[i];
						}
					}
					// facy se vzjemn ozna za sousedy, pokud
					// le odvrcen od sebe (objekt je konvex)
				}
			}
		}
	}
	// hled sousedn facy
}

extern unsigned __int32 *buffer;

void Project_Silhouette(Vector v_light, Object *p_object,
		Matrix *p_projection, unsigned __int32 *p_video)
{
	int n_edge_a, n_edge_b;
	Vertex v_end[3];
	int b_facing;
    Vertex v[3];
	Face face;
	Matrix im;
	Plane pl;
    int i, j;
	float f;

	Matrix_Inverse(p_projection, &im);

    face.vertex[0] = &v[0];
    face.vertex[1] = &v[1];
    face.vertex[2] = &v[2];
	// nastav face

	for(i = 0; i < p_object->n_face_num; i ++) {
		p_object->face[i].n_used = ((v_light.x * p_object->face[i].normal.a +
			v_light.y * p_object->face[i].normal.b + p_object->face[i].normal.d +
			v_light.z * p_object->face[i].normal.c) < 0)? true : false;
	}
	// nastav n_used na jedniku pro facy, kter jsou ze svtla vidt

	for(i = 0; i < p_object->n_face_num; i ++) {
		if(!p_object->face[i].n_used)
			continue;
		// stny vrhaj jen viditeln facy

		memcpy(face.vertex[2], p_object->face[i].vertex[0], sizeof(Vertex));
		memcpy(face.vertex[1], p_object->face[i].vertex[1], sizeof(Vertex));
		memcpy(face.vertex[0], p_object->face[i].vertex[2], sizeof(Vertex));
		//
		b_facing = (p_object->face[i].normal.c * im[3][2] +
			p_object->face[i].normal.a * im[3][0] +
			p_object->face[i].normal.b * im[3][1] + p_object->face[i].normal.d) < -0.1;
		//
		if(!b_facing) {
			memcpy(face.vertex[2], p_object->face[i].vertex[1], sizeof(Vertex));
			memcpy(face.vertex[1], p_object->face[i].vertex[0], sizeof(Vertex));
		}
		StencilFace(&face, p_projection, p_video, b_facing, false);
		// kresl elo stnu (jen bez S-Bufferu)

		for(j = 0; j < 3; j ++) {
			v_end[j].x = p_object->face[i].vertex[j]->x - v_light.x;
			v_end[j].y = p_object->face[i].vertex[j]->y - v_light.y;
			v_end[j].z = p_object->face[i].vertex[j]->z - v_light.z;
			//
			f = 256 / _Len(v_end[j]);
			v_end[j].x = p_object->face[i].vertex[j]->x + v_end[j].x * f;
			v_end[j].y = p_object->face[i].vertex[j]->y + v_end[j].y * f;
			v_end[j].z = p_object->face[i].vertex[j]->z + v_end[j].z * f;
		}
		// spot vzdlenou projekci facu (256 bod daleko)

		memcpy(face.vertex[0], &v_end[0], sizeof(Vertex));
		memcpy(face.vertex[1], &v_end[1], sizeof(Vertex));
		memcpy(face.vertex[2], &v_end[2], sizeof(Vertex));
		//
		pl = Make_Plane(&face);
		b_facing = (pl.c * im[3][2] + pl.a * im[3][0] +
			pl.b * im[3][1] + pl.d) > 0.1;
		//
		if(!b_facing) {
			memcpy(face.vertex[0], &v_end[1], sizeof(Vertex));
			memcpy(face.vertex[1], &v_end[0], sizeof(Vertex));
		}
		StencilFace(&face, p_projection, p_video, b_facing, true);
		// kresl posunutou face (tm vytvo zadn stnu volume)

		for(n_edge_a = 0; n_edge_a < 3; n_edge_a ++) {
			if(p_object->face[i].n_neighbour[n_edge_a] == NULL ||
			   p_object->face[i].n_neighbour[n_edge_a]->n_used == 0) {	
				n_edge_b = (1 + n_edge_a) % 3;
				// indexy vertex hrany (a, b)

				memcpy(face.vertex[0], &v_end[n_edge_a], sizeof(Vertex));
				memcpy(face.vertex[1], p_object->face[i].vertex[n_edge_a], sizeof(Vertex));
				memcpy(face.vertex[2], p_object->face[i].vertex[n_edge_b], sizeof(Vertex));

//#define __DISPLAY_VOLUME
//#define __DISPLAY_ACTIVE_EDGES

#ifdef __DISPLAY_VOLUME
				Line(*p_projection, &v_end[n_edge_a], &v_end[n_edge_b], buffer, 0x96ff64);
				Line(*p_projection, &v_end[n_edge_a], p_object->face[i].vertex[n_edge_a], buffer, 0x96ff64);
				Line(*p_projection, p_object->face[i].vertex[n_edge_b], &v_end[n_edge_b], buffer, 0x96ff64);
#endif
#ifdef __DISPLAY_ACTIVE_EDGES
				if(p_object->face[i].n_neighbour[n_edge_a] == NULL) {
					Line(*p_projection, p_object->face[i].vertex[n_edge_b],
						p_object->face[i].vertex[n_edge_a], buffer, 0xffc0c0);
				} else {
					Line(*p_projection, p_object->face[i].vertex[n_edge_b],
						p_object->face[i].vertex[n_edge_a], buffer, 0x96ff64);
				}
#endif
				// zobrazuje stnovou geometrii

				pl = Make_Plane(&face);
				b_facing = (pl.c * im[3][2] + pl.a * im[3][0] +
					        pl.b * im[3][1] + pl.d) > 0.1;
				if(b_facing) {
					StencilFace(&face, p_projection, p_video, b_facing, true);
					//
					memcpy(face.vertex[0], p_object->face[i].vertex[n_edge_b], sizeof(Vertex));
					memcpy(face.vertex[1], &v_end[n_edge_b], sizeof(Vertex));
					memcpy(face.vertex[2], &v_end[n_edge_a], sizeof(Vertex));
					StencilFace(&face, p_projection, p_video, b_facing, true);
				} else {
					memcpy(face.vertex[2], p_object->face[i].vertex[n_edge_a], sizeof(Vertex));
					memcpy(face.vertex[1], p_object->face[i].vertex[n_edge_b], sizeof(Vertex));
					StencilFace(&face, p_projection, p_video, b_facing, true);
					//
					memcpy(face.vertex[1], p_object->face[i].vertex[n_edge_b], sizeof(Vertex));
					memcpy(face.vertex[0], &v_end[n_edge_b], sizeof(Vertex));
					memcpy(face.vertex[2], &v_end[n_edge_a], sizeof(Vertex));
					StencilFace(&face, p_projection, p_video, b_facing, true);
				}
				// kresl facy
			}

		}
	}
}
/*
void DrawObject(Matrix *m_object, Matrix *m_camera,
    Object *p_object, Material *p_material)
{
    Face *cur_face, face;
	int n_material_id;
    float tx, ty, tz;
    float x, y, z;
    Polygon poly;
    Vertex v[3];
    int i, j;

    face.vertex[0] = &v[0];
    face.vertex[1] = &v[1];
    face.vertex[2] = &v[2];
    // nastav vertexy pro nai doasnou face

    Matrix_Inverse(m_camera, &m);
    Matrix_Multiply(&m, m_object, &m);
    Matrix_Inverse(&m, &im);
    // spot transforman matice

    for(i = 0; i < p_object->n_face_num; i ++){
        if((p_object->face[i].normal.c * im[3][2] +
           p_object->face[i].normal.a * im[3][0] +
           p_object->face[i].normal.b * im[3][1] +
           p_object->face[i].normal.d) > 0.1)
            continue;
        // byckface culling

        cur_face = &p_object->face[i];
		n_material_id = cur_face->n_material;
        for(j = 0; j < 3; j ++){
            x = cur_face->vertex[j]->x;
            y = cur_face->vertex[j]->y;
            z = cur_face->vertex[j]->z;
            //
            tx = x * m[0][0] + y * m[1][0] + 
                 z * m[2][0] +     m[3][0];
            ty = x * m[0][1] + y * m[1][1] + 
                 z * m[2][1] +     m[3][1];
            tz = x * m[0][2] + y * m[1][2] + 
                 z * m[2][2] +     m[3][2];
            //
            face.vertex[j]->x = tx;
            face.vertex[j]->y = ty;
            face.vertex[j]->z = tz;
            //
            face.vertex[j]->u = cur_face->vertex[j]->u;
            face.vertex[j]->v = cur_face->vertex[j]->v;
        }

        poly = Clip_Face(&face, &pyramid);
        // proene face pyramidou

        if(poly.n_vertex_num) { // pokud je face vidt, bude to nenulov slo
            for(j = 0; j < poly.n_vertex_num; j ++) {
                if(poly.vertex[j].z < 0.1) {
                    poly.n_vertex_num = 0;
                    break;
                }
                // pokud je z - souadnice moc mal, zahodme polygon

                poly.vertex[j].x = (n_Width / 2) + (poly.vertex[j].x *
                    z_delta) / poly.vertex[j].z;
                poly.vertex[j].y = (n_Height / 2) + (poly.vertex[j].y *
                    z_delta) / poly.vertex[j].z;
                // perspektivn korekce
            }

            DrawPolygon(&poly, p_material[n_material_id].texture);
            // vykresl polygon na obrazovku
        }
    }
}*/